import pytest
from rdflib import SH
from rdflib import Graph
from rdflib import Namespace

from dsp_tools.commands.validate_data.models.validation import CardinalityValidationResult
from dsp_tools.commands.validate_data.models.validation import ContentValidationResult

ONTO = Namespace("http://0.0.0.0:3333/ontology/9999/onto/v2#")
DATA = Namespace("http://data/")
KNORA_API = Namespace("http://api.knora.org/ontology/knora-api/v2#")


VALIDATION_PREFIXES = """
    @prefix onto: <http://0.0.0.0:3333/ontology/9999/onto/v2#> .
    @prefix sh: <http://www.w3.org/ns/shacl#> .
    @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
    @prefix api-shapes: <http://api.knora.org/ontology/knora-api/shapes/v2#> .
    @prefix knora-api: <http://api.knora.org/ontology/knora-api/v2#> .
    """


@pytest.fixture
def min_count_violation() -> Graph:
    gstr = f"""{VALIDATION_PREFIXES}
    [ a sh:ValidationResult ;
            sh:focusNode <http://data/id_min_card> ;
            sh:resultMessage "1-n" ;
            sh:resultPath onto:testGeoname ;
            sh:resultSeverity sh:Violation ;
            sh:sourceConstraintComponent sh:MinCountConstraintComponent ;
            sh:sourceShape [ ] ] .
    """
    g = Graph()
    g.parse(data=gstr, format="ttl")
    return g


@pytest.fixture
def data_min_count_violation() -> Graph:
    gstr = "<http://data/id_min_card> a <http://0.0.0.0:3333/ontology/9999/onto/v2#ClassMixedCard> ."
    g = Graph()
    g.parse(data=gstr, format="ttl")
    return g


@pytest.fixture
def graph_value_type_mismatch() -> Graph:
    gstr = f'''{VALIDATION_PREFIXES}
    [] a sh:ValidationReport ;
    sh:result _:bn_value_type_mismatch .

    [] a sh:ValidationReport ;
    sh:conforms false ;
    sh:result [ 
            a sh:ValidationResult ;
            sh:detail _:bn_value_type_mismatch ;
            sh:focusNode <http://data/value_type_mismatch> ;
            sh:resultMessage """Value does not have shape
                 <http://api.knora.org/ontology/knora-api/shapes/v2#ColorValue_ClassShape>""" ;
            sh:resultPath onto:testColor ;
            sh:resultSeverity sh:Violation ;
            sh:sourceConstraintComponent sh:NodeConstraintComponent ;
            sh:sourceShape onto:testColor_PropShape ;
            sh:value <http://data/value-iri-value_type_mismatch> 
            ] .

    _:bn_value_type_mismatch a sh:ValidationResult ;
    sh:focusNode <http://data/value-iri> ;
    sh:resultMessage "ColorValue" ;
    sh:resultSeverity sh:Violation ;
    sh:sourceConstraintComponent sh:ClassConstraintComponent ;
    sh:sourceShape <http://api.knora.org/ontology/knora-api/shapes/v2#ColorValue_ClassShape> ;
    sh:value <http://data/value-iri-value_type_mismatch> .
    '''
    g = Graph()
    g.parse(data=gstr, format="ttl")
    return g


@pytest.fixture
def data_value_type_mismatch() -> Graph:
    gstr = """
    <http://data/value_type_mismatch> a <http://0.0.0.0:3333/ontology/9999/onto/v2#ClassWithEverything> .
    <http://data/value-iri-value_type_mismatch> a <http://api.knora.org/ontology/knora-api/v2#TextValue> .
    """
    g = Graph()
    g.parse(data=gstr, format="ttl")
    return g


@pytest.fixture
def graph_wrong_regex_content() -> Graph:
    gstr = f'''{VALIDATION_PREFIXES}
    [] a sh:ValidationReport ;
    sh:result _:bn_geoname_not_number .

    [] a sh:ValidationReport ;
    sh:conforms false ;
    sh:result [ a sh:ValidationResult ;
            sh:detail _:bn_geoname_not_number ;
            sh:focusNode <http://data/geoname_not_number> ;
            sh:resultMessage """Value does not have shape 
                                        <http://api.knora.org/ontology/knora-api/shapes/v2#GeonameValue_ClassShape>""" ;
            sh:resultPath onto:testGeoname ;
            sh:resultSeverity sh:Violation ;
            sh:sourceConstraintComponent sh:NodeConstraintComponent ;
            sh:sourceShape onto:testGeoname_PropShape ;
            sh:value <http://data/value-iri-geoname_not_number> ] .

    _:bn_geoname_not_number a sh:ValidationResult ;
    sh:focusNode <http://data/value-iri-geoname_not_number> ;
    sh:resultMessage "The value must be a valid geoname code" ;
    sh:resultPath knora-api:geonameValueAsGeonameCode ;
    sh:resultSeverity sh:Violation ;
    sh:sourceConstraintComponent sh:PatternConstraintComponent ;
    sh:sourceShape api-shapes:geonameValueAsGeonameCode_Shape ;
    sh:value "this-is-not-a-valid-code" .
    '''
    g = Graph()
    g.parse(data=gstr, format="ttl")
    return g


@pytest.fixture
def data_wrong_regex_content() -> Graph:
    gstr = """
    <http://data/geoname_not_number> a <http://0.0.0.0:3333/ontology/9999/onto/v2#ClassWithEverything> .
    <http://data/value-iri-geoname_not_number> a <http://api.knora.org/ontology/knora-api/v2#GeonameValue> .
    """
    g = Graph()
    g.parse(data=gstr, format="ttl")
    return g


@pytest.fixture
def violation_min_card() -> CardinalityValidationResult:
    return CardinalityValidationResult(
        source_constraint_component=SH.MinCountConstraintComponent,
        res_iri=DATA.id_min_card,
        res_class=ONTO.ClassMixedCard,
        property=ONTO.testGeoname,
        results_message="1-n",
    )


@pytest.fixture
def violation_max_card() -> CardinalityValidationResult:
    return CardinalityValidationResult(
        source_constraint_component=SH.MaxCountConstraintComponent,
        res_iri=DATA.id_max_card,
        res_class=ONTO.ClassMixedCard,
        property=ONTO.testDecimalSimpleText,
        results_message="0-1",
    )


@pytest.fixture
def violation_closed() -> CardinalityValidationResult:
    return CardinalityValidationResult(
        source_constraint_component=SH.ClosedConstraintComponent,
        res_iri=DATA.id_closed_constraint,
        res_class=ONTO.CardOneResource,
        property=ONTO.testIntegerSimpleText,
        results_message="Predicate ns3:testIntegerSimpleText is not allowed (closed shape)",
    )


@pytest.fixture
def violation_value_type() -> ContentValidationResult:
    return ContentValidationResult(
        source_constraint_component=SH.NodeConstraintComponent,
        res_iri=DATA.id_2,
        res_class=ONTO.ClassWithEverything,
        property=ONTO.testColor,
        results_message="ColorValue",
        value_type=KNORA_API.TextValue,
        detail_bn_component=SH.ClassConstraintComponent,
        value=None,
    )


@pytest.fixture
def violation_unknown_content() -> ContentValidationResult:
    return ContentValidationResult(
        source_constraint_component=SH.UniqueLangConstraintComponent,
        res_iri=DATA.id,
        res_class=ONTO.ClassMixedCard,
        property=ONTO.testIntegerSimpleText,
        results_message="This is a constraint that is not checked in the data and should never appear.",
        detail_bn_component=SH.AndConstraintComponent,
        value=None,
    )


@pytest.fixture
def violation_regex() -> ContentValidationResult:
    return ContentValidationResult(
        source_constraint_component=SH.NodeConstraintComponent,
        res_iri=DATA.geoname_not_number,
        res_class=ONTO.ClassWithEverything,
        property=ONTO.testGeoname,
        results_message="The value must be a valid geoname code",
        value_type=KNORA_API.GeonameValue,
        detail_bn_component=SH.PatternConstraintComponent,
        value="this-is-not-a-valid-code",
    )



from copy import deepcopy
from dataclasses import dataclass
from unittest.mock import Mock

import pytest
from lxml import etree
from requests import Response

import dsp_tools.commands.xmlupload.models.upload_clients
from dsp_tools.commands.xmlupload import xmlupload
from dsp_tools.commands.xmlupload.iri_resolver import IriResolver
from dsp_tools.commands.xmlupload.models.deserialise.xmlresource import XMLResource
from dsp_tools.commands.xmlupload.models.ingest import AssetClient
from dsp_tools.commands.xmlupload.models.ingest import DspIngestClientLive
from dsp_tools.commands.xmlupload.models.upload_clients import UploadClients
from dsp_tools.commands.xmlupload.models.upload_state import UploadState
from dsp_tools.commands.xmlupload.project_client import ProjectInfo
from dsp_tools.commands.xmlupload.stash.stash_models import LinkValueStash
from dsp_tools.commands.xmlupload.stash.stash_models import LinkValueStashItem
from dsp_tools.commands.xmlupload.stash.stash_models import Stash
from dsp_tools.commands.xmlupload.upload_config import UploadConfig
from dsp_tools.models.custom_warnings import DspToolsUserWarning
from dsp_tools.models.exceptions import PermanentTimeOutError
from dsp_tools.models.exceptions import XmlUploadInterruptedError
from dsp_tools.utils.connection import Connection
from dsp_tools.utils.connection_live import ConnectionLive


@pytest.fixture
def ingest_client_mock():  # type: ignore[no-untyped-def]
    return Mock(spec_set=AssetClient)


class ListClientMock:
    def get_list_node_id_to_iri_lookup(self) -> dict[str, str]:
        return {}


@dataclass
class ProjectClientStub:
    """Stub class for ProjectClient."""

    con: Connection
    shortcode: str
    project_info: ProjectInfo | None

    def get_project_iri(self) -> str:
        return "https://admin.test.dasch.swiss/project/MsOaiQkcQ7-QPxsYBKckfQ"

    def get_ontology_iris(self) -> list[str]:
        raise NotImplementedError("get_project_iri not implemented")

    def get_ontology_name_dict(self) -> dict[str, str]:
        return {}

    def get_ontology_iri_dict(self) -> dict[str, str]:
        raise NotImplementedError("get_project_iri not implemented")


def test_one_resource_without_links(ingest_client_mock: AssetClient) -> None:
    xml_strings = [
        """
        <resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id">
            <text-prop name=":hasSimpleText"><text encoding="utf8">foo_1 text</text></text-prop>
        </resource>
        """,
    ]
    xml_resources = [XMLResource.from_node(etree.fromstring(xml_str), "my_onto") for xml_str in xml_strings]
    upload_state = UploadState(xml_resources, None, UploadConfig(), {})
    con = Mock(spec_set=ConnectionLive)
    post_responses = [{"@id": "foo_1_iri", "rdfs:label": "foo_1_label"}]
    con.post = Mock(side_effect=post_responses)
    project_client = ProjectClientStub(con, "1234", None)
    clients = dsp_tools.commands.xmlupload.models.upload_clients.UploadClients(
        ingest_client_mock, project_client, ListClientMock()
    )

    xmlupload._upload_resources(clients, upload_state)

    assert len(con.post.call_args_list) == len(post_responses)
    match con.post.call_args_list[0].kwargs:
        case {
            "route": "/v2/resources",
            "data": {
                "@type": "my_onto:foo_1_type",
                "rdfs:label": "foo_1_label",
                "knora-api:attachedToProject": {"@id": "https://admin.test.dasch.swiss/project/MsOaiQkcQ7-QPxsYBKckfQ"},
                "@context": dict(),
                "my_onto:hasSimpleText": [{"@type": "knora-api:TextValue", "knora-api:valueAsString": "foo_1 text"}],
            },
        }:
            assert True
        case _:
            pytest.fail("POST request was not sent correctly")
    assert not upload_state.pending_resources
    assert not upload_state.failed_uploads
    assert upload_state.iri_resolver.lookup == {"foo_1_id": "foo_1_iri"}
    assert not upload_state.pending_stash


def test_one_resource_with_link_to_existing_resource(ingest_client_mock: AssetClient) -> None:
    xml_strings = [
        """
        <resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id">
            <resptr-prop name=":hasCustomLink"><resptr>foo_2_id</resptr></resptr-prop>
        </resource>
        """,
    ]
    xml_resources = [XMLResource.from_node(etree.fromstring(xml_str), "my_onto") for xml_str in xml_strings]
    upload_state = UploadState(xml_resources, None, UploadConfig(), {}, [], IriResolver({"foo_2_id": "foo_2_iri"}))
    con = Mock(spec_set=ConnectionLive)
    post_responses = [{"@id": "foo_1_iri", "rdfs:label": "foo_1_label"}]
    con.post = Mock(side_effect=post_responses)
    project_client = ProjectClientStub(con, "1234", None)
    clients = dsp_tools.commands.xmlupload.models.upload_clients.UploadClients(
        ingest_client_mock, project_client, ListClientMock()
    )

    xmlupload._upload_resources(clients, upload_state)

    assert len(con.post.call_args_list) == len(post_responses)
    match con.post.call_args_list[0].kwargs:
        case {
            "route": "/v2/resources",
            "data": {
                "@type": "my_onto:foo_1_type",
                "rdfs:label": "foo_1_label",
                "knora-api:attachedToProject": {"@id": "https://admin.test.dasch.swiss/project/MsOaiQkcQ7-QPxsYBKckfQ"},
                "@context": dict(),
                "my_onto:hasCustomLinkValue": [
                    {"@type": "knora-api:LinkValue", "knora-api:linkValueHasTargetIri": {"@id": "foo_2_iri"}}
                ],
            },
        }:
            assert True
        case _:
            pytest.fail("POST request was not sent correctly")
    assert not upload_state.pending_resources
    assert not upload_state.failed_uploads
    assert upload_state.iri_resolver.lookup == {"foo_1_id": "foo_1_iri", "foo_2_id": "foo_2_iri"}
    assert not upload_state.pending_stash


def test_2_resources_with_stash_interrupted_by_timeout(ingest_client_mock: AssetClient) -> None:
    _2_resources_with_stash_interrupted_by_error(PermanentTimeOutError(""), "PermanentTimeOutError", ingest_client_mock)


def test_2_resources_with_stash_interrupted_by_keyboard(ingest_client_mock: AssetClient) -> None:
    _2_resources_with_stash_interrupted_by_error(KeyboardInterrupt(), "KeyboardInterrupt", ingest_client_mock)


def _2_resources_with_stash_interrupted_by_error(
    err_to_interrupt_with: BaseException, err_as_str: str, ingest_client_mock: AssetClient
) -> None:
    xml_strings = [
        '<resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id"></resource>',
        '<resource label="foo_2_label" restype=":foo_2_type" id="foo_2_id"></resource>',
    ]
    xml_resources = [XMLResource.from_node(etree.fromstring(xml_str), "my_onto") for xml_str in xml_strings]
    link_val_stash_dict = {
        "foo_1_id": [LinkValueStashItem("foo_1_id", "my_onto:foo_1_type", "my_onto:hasCustomLink", "foo_2_id")],
        "foo_2_id": [LinkValueStashItem("foo_2_id", "my_onto:foo_2_type", "my_onto:hasCustomLink", "foo_1_id")],
    }
    stash = Stash(link_value_stash=LinkValueStash(link_val_stash_dict), standoff_stash=None)
    upload_state = UploadState(xml_resources.copy(), deepcopy(stash), UploadConfig(), {})
    con = Mock(spec_set=ConnectionLive)
    post_responses = [
        {"@id": "foo_1_iri", "rdfs:label": "foo_1_label"},
        err_to_interrupt_with,
    ]
    con.post = Mock(side_effect=post_responses)
    project_client = ProjectClientStub(con, "1234", None)
    xmlupload._handle_upload_error = Mock()
    clients = dsp_tools.commands.xmlupload.models.upload_clients.UploadClients(
        ingest_client_mock, project_client, ListClientMock()
    )

    with pytest.warns(DspToolsUserWarning):
        xmlupload._upload_resources(clients, upload_state)

    assert len(con.post.call_args_list) == len(post_responses)
    err_msg = (
        f"There was a {err_as_str} while trying to create resource 'foo_2_id'.\n"
        "It is unclear if the resource 'foo_2_id' was created successfully or not.\n"
        "Please check manually in the DSP-APP or DB.\n"
        "In case of successful creation, call 'resume-xmlupload' with the flag "
        "'--skip-first-resource' to prevent duplication.\n"
        "If not, a normal 'resume-xmlupload' can be started."
    )
    upload_state_expected = UploadState(
        xml_resources[1:], stash, UploadConfig(), {}, [], IriResolver({"foo_1_id": "foo_1_iri"})
    )
    xmlupload._handle_upload_error.assert_called_once_with(XmlUploadInterruptedError(err_msg), upload_state_expected)


def test_2_resources_with_stash(ingest_client_mock: AssetClient) -> None:
    xml_strings = [
        '<resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id"></resource>',
        '<resource label="foo_2_label" restype=":foo_2_type" id="foo_2_id"></resource>',
    ]
    xml_resources = [XMLResource.from_node(etree.fromstring(xml_str), "my_onto") for xml_str in xml_strings]
    link_val_stash_dict = {
        "foo_1_id": [LinkValueStashItem("foo_1_id", "my_onto:foo_1_type", "my_onto:hasCustomLink", "foo_2_id")],
        "foo_2_id": [LinkValueStashItem("foo_2_id", "my_onto:foo_2_type", "my_onto:hasCustomLink", "foo_1_id")],
    }
    stash = Stash(link_value_stash=LinkValueStash(link_val_stash_dict), standoff_stash=None)
    upload_state = UploadState(xml_resources.copy(), deepcopy(stash), UploadConfig(), {})
    con = Mock(spec_set=ConnectionLive)
    post_responses = [
        {"@id": "foo_1_iri", "rdfs:label": "foo_1_label"},
        {"@id": "foo_2_iri", "rdfs:label": "foo_2_label"},
        {},  # uploading a stash doesn't rely on a certain response
        {},  # uploading a stash doesn't rely on a certain response
    ]
    con.post = Mock(side_effect=post_responses)
    project_client = ProjectClientStub(con, "1234", None)
    clients = dsp_tools.commands.xmlupload.models.upload_clients.UploadClients(
        ingest_client_mock, project_client, ListClientMock()
    )

    xmlupload._upload_resources(clients, upload_state)

    assert len(con.post.call_args_list) == len(post_responses)
    match con.post.call_args_list[2].kwargs:
        case {
            "route": "/v2/values",
            "data": {
                "@type": "my_onto:foo_1_type",
                "@id": "foo_1_iri",
                "@context": dict(),
                "my_onto:hasCustomLinkValue": {
                    "@type": "knora-api:LinkValue",
                    "knora-api:linkValueHasTargetIri": {"@id": "foo_2_iri"},
                },
            },
        }:
            assert True
        case _:
            pytest.fail("POST request was not sent correctly")
    assert not upload_state.pending_resources
    assert not upload_state.failed_uploads
    assert upload_state.iri_resolver.lookup == {"foo_1_id": "foo_1_iri", "foo_2_id": "foo_2_iri"}
    assert not upload_state.pending_stash or upload_state.pending_stash.is_empty()


def test_5_resources_with_stash_and_interrupt_after_2(ingest_client_mock: AssetClient) -> None:
    xml_strings = [
        '<resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id"></resource>',
        '<resource label="foo_2_label" restype=":foo_2_type" id="foo_2_id"></resource>',
        '<resource label="foo_3_label" restype=":foo_3_type" id="foo_3_id"></resource>',
        '<resource label="foo_4_label" restype=":foo_4_type" id="foo_4_id"></resource>',
        '<resource label="foo_5_label" restype=":foo_5_type" id="foo_5_id"></resource>',
    ]
    xml_resources = [XMLResource.from_node(etree.fromstring(xml_str), "my_onto") for xml_str in xml_strings]
    link_val_stash_dict = {
        "foo_1_id": [LinkValueStashItem("foo_1_id", "my_onto:foo_1_type", "my_onto:hasCustomLink", "foo_2_id")],
        "foo_2_id": [LinkValueStashItem("foo_2_id", "my_onto:foo_2_type", "my_onto:hasCustomLink", "foo_1_id")],
    }
    stash = Stash(link_value_stash=LinkValueStash(link_val_stash_dict), standoff_stash=None)
    upload_config = UploadConfig(interrupt_after=2)
    upload_state = UploadState(xml_resources.copy(), deepcopy(stash), upload_config, {})
    con = Mock(spec_set=ConnectionLive)
    post_responses = [
        {"@id": "foo_1_iri", "rdfs:label": "foo_1_label"},
        {"@id": "foo_2_iri", "rdfs:label": "foo_2_label"},
        {"@id": "foo_3_iri", "rdfs:label": "foo_3_label"},
        {"@id": "foo_4_iri", "rdfs:label": "foo_4_label"},
        {"@id": "foo_5_iri", "rdfs:label": "foo_5_label"},
        {},  # uploading a stash doesn't rely on a certain response
        {},  # uploading a stash doesn't rely on a certain response
    ]
    con.post = Mock(side_effect=post_responses)
    project_client = ProjectClientStub(con, "1234", None)
    xmlupload._handle_upload_error = Mock()
    err_msg = "Interrupted: Maximum number of resources was reached (2)"

    client = dsp_tools.commands.xmlupload.models.upload_clients.UploadClients(
        ingest_client_mock, project_client, ListClientMock()
    )

    xmlupload._upload_resources(client, upload_state)
    iri_resolver_expected = IriResolver({"foo_1_id": "foo_1_iri", "foo_2_id": "foo_2_iri"})
    upload_state_expected = UploadState(xml_resources[2:], stash, upload_config, {}, [], iri_resolver_expected)
    xmlupload._handle_upload_error.assert_called_once_with(XmlUploadInterruptedError(err_msg), upload_state_expected)

    xmlupload._handle_upload_error = Mock()
    xmlupload._upload_resources(client, upload_state)
    iri_resolver_expected.lookup.update({"foo_3_id": "foo_3_iri", "foo_4_id": "foo_4_iri"})
    upload_state_expected = UploadState(xml_resources[4:], stash, upload_config, {}, [], iri_resolver_expected)
    xmlupload._handle_upload_error.assert_called_once_with(XmlUploadInterruptedError(err_msg), upload_state_expected)

    xmlupload._handle_upload_error = Mock()
    xmlupload._upload_resources(client, upload_state)
    iri_resolver_expected.lookup.update({"foo_5_id": "foo_5_iri"})
    empty_stash = Stash(standoff_stash=None, link_value_stash=LinkValueStash({}))
    upload_state_expected = UploadState([], empty_stash, upload_config, {}, [], iri_resolver_expected)
    xmlupload._handle_upload_error.assert_not_called()
    assert upload_state == upload_state_expected


def test_6_resources_with_stash_and_interrupt_after_2(ingest_client_mock: AssetClient) -> None:
    xml_strings = [
        '<resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id"></resource>',
        '<resource label="foo_2_label" restype=":foo_2_type" id="foo_2_id"></resource>',
        '<resource label="foo_3_label" restype=":foo_3_type" id="foo_3_id"></resource>',
        '<resource label="foo_4_label" restype=":foo_4_type" id="foo_4_id"></resource>',
        '<resource label="foo_5_label" restype=":foo_5_type" id="foo_5_id"></resource>',
        '<resource label="foo_6_label" restype=":foo_6_type" id="foo_6_id"></resource>',
    ]
    xml_resources = [XMLResource.from_node(etree.fromstring(xml_str), "my_onto") for xml_str in xml_strings]
    link_val_stash_dict = {
        "foo_1_id": [LinkValueStashItem("foo_1_id", "my_onto:foo_1_type", "my_onto:hasCustomLink", "foo_2_id")],
        "foo_2_id": [LinkValueStashItem("foo_2_id", "my_onto:foo_2_type", "my_onto:hasCustomLink", "foo_1_id")],
    }
    stash = Stash(link_value_stash=LinkValueStash(link_val_stash_dict), standoff_stash=None)
    upload_config = UploadConfig(interrupt_after=2)
    upload_state = UploadState(xml_resources.copy(), deepcopy(stash), upload_config, {})
    con = Mock(spec_set=ConnectionLive)
    post_responses = [
        {"@id": "foo_1_iri", "rdfs:label": "foo_1_label"},
        {"@id": "foo_2_iri", "rdfs:label": "foo_2_label"},
        {"@id": "foo_3_iri", "rdfs:label": "foo_3_label"},
        {"@id": "foo_4_iri", "rdfs:label": "foo_4_label"},
        {"@id": "foo_5_iri", "rdfs:label": "foo_5_label"},
        {"@id": "foo_6_iri", "rdfs:label": "foo_6_label"},
        {},  # uploading a stash doesn't rely on a certain response
        {},  # uploading a stash doesn't rely on a certain response
    ]
    con.post = Mock(side_effect=post_responses)
    project_client = ProjectClientStub(con, "1234", None)
    xmlupload._handle_upload_error = Mock()
    err_msg = "Interrupted: Maximum number of resources was reached (2)"

    client = dsp_tools.commands.xmlupload.models.upload_clients.UploadClients(
        ingest_client_mock, project_client, ListClientMock()
    )

    xmlupload._upload_resources(client, upload_state)
    iri_resolver_expected = IriResolver({"foo_1_id": "foo_1_iri", "foo_2_id": "foo_2_iri"})
    upload_state_expected = UploadState(xml_resources[2:], stash, upload_config, {}, [], iri_resolver_expected)
    xmlupload._handle_upload_error.assert_called_once_with(XmlUploadInterruptedError(err_msg), upload_state_expected)

    xmlupload._handle_upload_error = Mock()
    xmlupload._upload_resources(client, upload_state)
    iri_resolver_expected.lookup.update({"foo_3_id": "foo_3_iri", "foo_4_id": "foo_4_iri"})
    upload_state_expected = UploadState(xml_resources[4:], stash, upload_config, {}, [], iri_resolver_expected)
    xmlupload._handle_upload_error.assert_called_once_with(XmlUploadInterruptedError(err_msg), upload_state_expected)

    xmlupload._handle_upload_error = Mock()
    xmlupload._upload_resources(client, upload_state)
    iri_resolver_expected.lookup.update({"foo_5_id": "foo_5_iri", "foo_6_id": "foo_6_iri"})
    upload_state_expected = UploadState([], stash, upload_config, {}, [], iri_resolver_expected)
    xmlupload._handle_upload_error.assert_called_once_with(XmlUploadInterruptedError(err_msg), upload_state_expected)

    xmlupload._handle_upload_error = Mock()
    xmlupload._upload_resources(client, upload_state)
    empty_stash = Stash(standoff_stash=None, link_value_stash=LinkValueStash({}))
    upload_state_expected = UploadState([], empty_stash, upload_config, {}, [], iri_resolver_expected)
    xmlupload._handle_upload_error.assert_not_called()
    assert upload_state == upload_state_expected


def test_logging(caplog: pytest.LogCaptureFixture, ingest_client_mock: AssetClient) -> None:
    xml_strings = [
        '<resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id"></resource>',
        '<resource label="foo_2_label" restype=":foo_2_type" id="foo_2_id"></resource>',
        '<resource label="foo_3_label" restype=":foo_3_type" id="foo_3_id"></resource>',
        '<resource label="foo_4_label" restype=":foo_4_type" id="foo_4_id"></resource>',
        '<resource label="foo_5_label" restype=":foo_5_type" id="foo_5_id"></resource>',
    ]
    xml_resources = [XMLResource.from_node(etree.fromstring(xml_str), "my_onto") for xml_str in xml_strings]
    link_val_stash_dict = {
        "foo_1_id": [LinkValueStashItem("foo_1_id", "my_onto:foo_1_type", "my_onto:hasCustomLink", "foo_2_id")],
        "foo_2_id": [LinkValueStashItem("foo_2_id", "my_onto:foo_2_type", "my_onto:hasCustomLink", "foo_1_id")],
    }
    stash = Stash(link_value_stash=LinkValueStash(link_val_stash_dict), standoff_stash=None)
    upload_config = UploadConfig(interrupt_after=2)
    upload_state = UploadState(xml_resources.copy(), deepcopy(stash), upload_config, {})
    con = Mock(spec_set=ConnectionLive)
    post_responses = [
        {"@id": "foo_1_iri", "rdfs:label": "foo_1_label"},
        {"@id": "foo_2_iri", "rdfs:label": "foo_2_label"},
        {"@id": "foo_3_iri", "rdfs:label": "foo_3_label"},
        {"@id": "foo_4_iri", "rdfs:label": "foo_4_label"},
        {"@id": "foo_5_iri", "rdfs:label": "foo_5_label"},
        {},  # uploading a stash doesn't rely on a certain response
        {},  # uploading a stash doesn't rely on a certain response
    ]
    con.post = Mock(side_effect=post_responses)
    project_client = ProjectClientStub(con, "1234", None)
    xmlupload._handle_upload_error = Mock()
    clients = dsp_tools.commands.xmlupload.models.upload_clients.UploadClients(
        ingest_client_mock, project_client, ListClientMock()
    )

    xmlupload._upload_resources(clients, upload_state)
    assert caplog.records[1].message == "Created resource 1/5: 'foo_1_label' (ID: 'foo_1_id', IRI: 'foo_1_iri')"
    assert caplog.records[3].message == "Created resource 2/5: 'foo_2_label' (ID: 'foo_2_id', IRI: 'foo_2_iri')"
    caplog.clear()

    xmlupload._upload_resources(clients, upload_state)
    assert caplog.records[1].message == "Created resource 3/5: 'foo_3_label' (ID: 'foo_3_id', IRI: 'foo_3_iri')"
    assert caplog.records[3].message == "Created resource 4/5: 'foo_4_label' (ID: 'foo_4_id', IRI: 'foo_4_iri')"
    caplog.clear()

    xmlupload._upload_resources(clients, upload_state)
    assert caplog.records[1].message == "Created resource 5/5: 'foo_5_label' (ID: 'foo_5_id', IRI: 'foo_5_iri')"
    assert caplog.records[3].message == "  Upload resptrs of resource 'foo_1_id'..."
    assert caplog.records[5].message == "  Upload resptrs of resource 'foo_2_id'..."
    caplog.clear()


def test_post_requests(ingest_client_mock: AssetClient) -> None:
    xml_strings = [
        '<resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id"></resource>',
        '<resource label="foo_2_label" restype=":foo_2_type" id="foo_2_id"></resource>',
        '<resource label="foo_3_label" restype=":foo_3_type" id="foo_3_id"></resource>',
        '<resource label="foo_4_label" restype=":foo_4_type" id="foo_4_id"></resource>',
        '<resource label="foo_5_label" restype=":foo_5_type" id="foo_5_id"></resource>',
        '<resource label="foo_6_label" restype=":foo_6_type" id="foo_6_id"></resource>',
    ]
    xml_resources = [XMLResource.from_node(etree.fromstring(xml_str), "my_onto") for xml_str in xml_strings]
    link_val_stash_dict = {
        "foo_1_id": [LinkValueStashItem("foo_1_id", "my_onto:foo_1_type", "my_onto:hasCustomLink", "foo_2_id")],
        "foo_2_id": [LinkValueStashItem("foo_2_id", "my_onto:foo_2_type", "my_onto:hasCustomLink", "foo_1_id")],
    }
    stash = Stash(link_value_stash=LinkValueStash(link_val_stash_dict), standoff_stash=None)
    upload_config = UploadConfig(interrupt_after=2)
    upload_state = UploadState(xml_resources.copy(), deepcopy(stash), upload_config, {})
    con = Mock(spec_set=ConnectionLive)
    post_responses = [
        {"@id": "foo_1_iri", "rdfs:label": "foo_1_label"},
        {"@id": "foo_2_iri", "rdfs:label": "foo_2_label"},
        {"@id": "foo_3_iri", "rdfs:label": "foo_3_label"},
        {"@id": "foo_4_iri", "rdfs:label": "foo_4_label"},
        {"@id": "foo_5_iri", "rdfs:label": "foo_5_label"},
        {"@id": "foo_6_iri", "rdfs:label": "foo_6_label"},
        {},  # uploading a stash doesn't rely on a certain response
        {},  # uploading a stash doesn't rely on a certain response
    ]
    con.post = Mock(side_effect=post_responses)
    project_client = ProjectClientStub(con, "1234", None)
    xmlupload._handle_upload_error = Mock()
    clients = dsp_tools.commands.xmlupload.models.upload_clients.UploadClients(
        ingest_client_mock, project_client, ListClientMock()
    )

    xmlupload._upload_resources(clients, upload_state)
    xmlupload._upload_resources(clients, upload_state)
    xmlupload._upload_resources(clients, upload_state)
    xmlupload._upload_resources(clients, upload_state)
    assert len(con.post.call_args_list) == len(post_responses)


def test_interruption_if_resource_cannot_be_created_because_of_404() -> None:
    xml_strings = [
        '<resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id"></resource>',
        '<resource label="foo_2_label" restype=":foo_2_type" id="foo_2_id"></resource>',
    ]
    xml_resources = [XMLResource.from_node(etree.fromstring(xml_str), "my_onto") for xml_str in xml_strings]
    upload_state = UploadState(xml_resources.copy(), Stash(None, None), UploadConfig(), {}, [], IriResolver())
    con = ConnectionLive("foo")
    con._log_request = Mock()  # type: ignore[method-assign]
    con._log_response = Mock()  # type: ignore[method-assign]
    resp_404 = Response()
    resp_404.status_code = 404
    post_responses = [resp_404]
    con.session.request = Mock(side_effect=post_responses)  # type: ignore[method-assign]
    project_client = ProjectClientStub(con, "1234", None)
    xmlupload._handle_upload_error = Mock()
    ingest_client = DspIngestClientLive("", "", "1234", ".")

    xmlupload._upload_resources(UploadClients(ingest_client, project_client, ListClientMock()), upload_state)
    msg = (
        "Lost connection to DSP server, probably because the server is down. "
        "Please continue later with 'resume-xmlupload'. Reason for this failure: "
        "Permanently unable to execute the network action. "
    )
    assert len(xmlupload._handle_upload_error.call_args_list) == 1
    err_actual: XmlUploadInterruptedError = xmlupload._handle_upload_error.call_args_list[0].args[0]
    upload_state_actual: UploadState = xmlupload._handle_upload_error.call_args_list[0].args[1]
    assert msg in err_actual.message
    assert upload_state_actual == upload_state



import logging
from typing import Iterator

import pytest
from _pytest.logging import caplog as _caplog  # noqa: F401 (imported but unused)
from loguru import logger

from dsp_tools.utils.logger_config import logger_config


@pytest.fixture
def caplog(_caplog: pytest.LogCaptureFixture) -> Iterator[pytest.LogCaptureFixture]:  # noqa: F811 (redefinition)
    """
    The caplog fixture that comes shipped with pytest does not support loguru.
    This modified version can be used exactly like the builtin caplog fixture,
    which is documented at https://docs.pytest.org/en/latest/how-to/logging.html#caplog-fixture.
    Credits: https://www.youtube.com/watch?v=eFdVlyAGeZU

    Yields:
        pytest.LogCaptureFixture: The modified caplog fixture.
    """

    class PropagateHandler(logging.Handler):
        def emit(self, record: logging.LogRecord) -> None:
            logging.getLogger(record.name).handle(record)

    handler_id = logger.add(sink=PropagateHandler(), format="{message}", level="DEBUG")
    yield _caplog
    logger.remove(handler_id)


def pytest_sessionstart() -> None:
    """
    Called after the Session object has been created
    and before performing collection and entering the run test loop.
    See https://docs.pytest.org/en/latest/reference/reference.html#pytest.hookspec.pytest_sessionstart.
    """
    logger_config()



import pytest

from dsp_tools.models.exceptions import BaseError
from dsp_tools.utils.date_util import Calendar
from dsp_tools.utils.date_util import Date
from dsp_tools.utils.date_util import Era
from dsp_tools.utils.date_util import SingleDate
from dsp_tools.utils.date_util import _parse_single_date
from dsp_tools.utils.date_util import _split_date_string
from dsp_tools.utils.date_util import parse_date_string


class TestParseDateStringWithDateOnly:
    """Tests parsing a date string with only dates, no era or calendar information."""

    def test_start_date(self) -> None:
        date_string = "2020-01-01"
        result = parse_date_string(date_string)
        expected_start_and_end = SingleDate(era=Era.CE, year=2020, month=1, day=1)
        expected = Date(Calendar.GREGORIAN, expected_start_and_end, expected_start_and_end)
        assert result == expected

    def test_start_year(self) -> None:
        date_string = "2020"
        result = parse_date_string(date_string)
        expected_start_and_end = SingleDate(era=Era.CE, year=2020, month=None, day=None)
        expected = Date(Calendar.GREGORIAN, expected_start_and_end, expected_start_and_end)
        assert result == expected

    def test_start_year_and_month(self) -> None:
        date_string = "2020-01"
        result = parse_date_string(date_string)
        expected_start_and_end = SingleDate(era=Era.CE, year=2020, month=1, day=None)
        expected = Date(Calendar.GREGORIAN, expected_start_and_end, expected_start_and_end)
        assert result == expected

    def test_start_and_end_year(self) -> None:
        date_string = "2020:2021"
        result = parse_date_string(date_string)
        expected_start = SingleDate(era=Era.CE, year=2020, month=None, day=None)
        expected_end = SingleDate(era=Era.CE, year=2021, month=None, day=None)
        expected = Date(Calendar.GREGORIAN, expected_start, expected_end)
        assert result == expected

    def test_start_and_end_date(self) -> None:
        date_string = "2020-01-01:2021-02-02"
        result = parse_date_string(date_string)
        expected_start = SingleDate(era=Era.CE, year=2020, month=1, day=1)
        expected_end = SingleDate(era=Era.CE, year=2021, month=2, day=2)
        expected = Date(Calendar.GREGORIAN, expected_start, expected_end)
        assert result == expected


class TestParseDateStringWithEra:
    """Tests parsing a date string with era information."""

    def test_start_date_ce(self) -> None:
        date_string = "CE:2020-01-01"
        result = parse_date_string(date_string)
        expected_start_and_end = SingleDate(era=Era.CE, year=2020, month=1, day=1)
        expected = Date(Calendar.GREGORIAN, expected_start_and_end, expected_start_and_end)
        assert result == expected

    def test_start_date_bce(self) -> None:
        date_string = "BCE:2020-01-01"
        result = parse_date_string(date_string)
        expected_start_and_end = SingleDate(era=Era.BCE, year=2020, month=1, day=1)
        expected = Date(Calendar.GREGORIAN, expected_start_and_end, expected_start_and_end)
        assert result == expected

    def test_start_date_ad(self) -> None:
        date_string = "AD:2020-01-01"
        result = parse_date_string(date_string)
        expected_start_and_end = SingleDate(era=Era.AD, year=2020, month=1, day=1)
        expected = Date(Calendar.GREGORIAN, expected_start_and_end, expected_start_and_end)
        assert result == expected

    def test_start_date_bc(self) -> None:
        date_string = "BC:2020-01-01"
        result = parse_date_string(date_string)
        expected_start_and_end = SingleDate(era=Era.BC, year=2020, month=1, day=1)
        expected = Date(Calendar.GREGORIAN, expected_start_and_end, expected_start_and_end)
        assert result == expected

    def test_start_and_end_date_ce_on_start_only(self) -> None:
        date_string = "CE:2020-01-01:2021-02-02"
        result = parse_date_string(date_string)
        expected_start = SingleDate(era=Era.CE, year=2020, month=1, day=1)
        expected_end = SingleDate(era=Era.CE, year=2021, month=2, day=2)
        expected = Date(Calendar.GREGORIAN, expected_start, expected_end)
        assert result == expected

    def test_start_and_end_date_bce_on_start_only(self) -> None:
        date_string = "BCE:2021-01-01:2020-02-02"
        result = parse_date_string(date_string)
        expected_start = SingleDate(era=Era.BCE, year=2021, month=1, day=1)
        expected_end = SingleDate(era=Era.BCE, year=2020, month=2, day=2)
        expected = Date(Calendar.GREGORIAN, expected_start, expected_end)
        assert result == expected

    def test_start_and_end_date_ce_on_both(self) -> None:
        date_string = "CE:2020-01-01:CE:2021-02-02"
        result = parse_date_string(date_string)
        expected_start = SingleDate(era=Era.CE, year=2020, month=1, day=1)
        expected_end = SingleDate(era=Era.CE, year=2021, month=2, day=2)
        expected = Date(Calendar.GREGORIAN, expected_start, expected_end)
        assert result == expected

    def test_start_and_end_date_bce_on_both(self) -> None:
        date_string = "BCE:2020-01-01:BCE:2021-02-02"
        result = parse_date_string(date_string)
        expected_start = SingleDate(era=Era.BCE, year=2020, month=1, day=1)
        expected_end = SingleDate(era=Era.BCE, year=2021, month=2, day=2)
        expected = Date(Calendar.GREGORIAN, expected_start, expected_end)
        assert result == expected

    def test_dates_across_era_border(self) -> None:
        date_string = "BCE:2020-01-01:CE:2021-02-02"
        result = parse_date_string(date_string)
        expected_start = SingleDate(era=Era.BCE, year=2020, month=1, day=1)
        expected_end = SingleDate(era=Era.CE, year=2021, month=2, day=2)
        expected = Date(Calendar.GREGORIAN, expected_start, expected_end)
        assert result == expected


class TestParseDateStringWithCalendar:
    """Tests parsing a date string with calendar information."""

    def test_start_date_gregorian(self) -> None:
        date_string = "GREGORIAN:2020-01-01"
        result = parse_date_string(date_string)
        expected_start_and_end = SingleDate(era=Era.CE, year=2020, month=1, day=1)
        expected = Date(Calendar.GREGORIAN, expected_start_and_end, expected_start_and_end)
        assert result == expected

    def test_start_date_gregorian_with_era(self) -> None:
        date_string = "GREGORIAN:BCE:2020-01-01"
        result = parse_date_string(date_string)
        expected_start_and_end = SingleDate(era=Era.BCE, year=2020, month=1, day=1)
        expected = Date(Calendar.GREGORIAN, expected_start_and_end, expected_start_and_end)
        assert result == expected

    def test_start_date_julian(self) -> None:
        date_string = "JULIAN:2020-01-01"
        result = parse_date_string(date_string)
        expected_start_and_end = SingleDate(era=Era.CE, year=2020, month=1, day=1)
        expected = Date(Calendar.JULIAN, expected_start_and_end, expected_start_and_end)
        assert result == expected

    def test_start_date_julian_with_era(self) -> None:
        date_string = "JULIAN:BCE:2020-01-01"
        result = parse_date_string(date_string)
        expected_start_and_end = SingleDate(era=Era.BCE, year=2020, month=1, day=1)
        expected = Date(Calendar.JULIAN, expected_start_and_end, expected_start_and_end)
        assert result == expected

    def test_start_date_islamic(self) -> None:
        date_string = "ISLAMIC:2020-01-01"
        result = parse_date_string(date_string)
        expected_start_and_end = SingleDate(era=None, year=2020, month=1, day=1)
        expected = Date(Calendar.ISLAMIC, expected_start_and_end, expected_start_and_end)
        assert result == expected

    def test_not_allow_islamic_calendar_with_era(self) -> None:
        date_string = "ISLAMIC:BCE:2020-01-01"
        with pytest.raises(BaseError):
            parse_date_string(date_string)


class TestSplitDateString:
    """Tests splitting a date string into its components."""

    def test_split_full_date_string(self) -> None:
        date_string = "GREGORIAN:CE:2020-01-01:CE:2021-02-02"
        result = _split_date_string(date_string)
        expected = ("GREGORIAN", "CE", "2020-01-01", "CE", "2021-02-02")
        assert result == expected

    def test_split_start_date_only(self) -> None:
        date_string = "2020-01-01"
        result = _split_date_string(date_string)
        expected = (None, None, "2020-01-01", None, None)
        assert result == expected

    def test_split_start_and_end_date_only(self) -> None:
        date_string = "2020-01-01:2021-02-02"
        result = _split_date_string(date_string)
        expected = (None, None, "2020-01-01", None, "2021-02-02")
        assert result == expected

    def test_split_start_date_with_era(self) -> None:
        date_string = "CE:2020-01-01"
        result = _split_date_string(date_string)
        expected = (None, "CE", "2020-01-01", None, None)
        assert result == expected

    def test_split_start_and_end_date_with_era(self) -> None:
        date_string = "CE:2020-01-01:CE:2021-02-02"
        result = _split_date_string(date_string)
        expected = (None, "CE", "2020-01-01", "CE", "2021-02-02")
        assert result == expected

    def test_split_start_date_with_calendar(self) -> None:
        date_string = "GREGORIAN:2020-01-01"
        result = _split_date_string(date_string)
        expected = ("GREGORIAN", None, "2020-01-01", None, None)
        assert result == expected


class TestParseSingleDate:
    """Tests parsing a single date."""

    def test_parse_date_string(self) -> None:
        date_string = "2020-01-01"
        result = _parse_single_date(date_string, Era.CE)
        expected = SingleDate(era=Era.CE, year=2020, month=1, day=1)
        assert result == expected

    def test_parse_year_only(self) -> None:
        date_string = "2020"
        result = _parse_single_date(date_string, Era.CE)
        expected = SingleDate(era=Era.CE, year=2020, month=None, day=None)
        assert result == expected

    def test_parse_year_and_month(self) -> None:
        date_string = "2020-01"
        result = _parse_single_date(date_string, Era.CE)
        expected = SingleDate(era=Era.CE, year=2020, month=1, day=None)
        assert result == expected

    def test_parse_date_fewer_digits_for_month_and_day(self) -> None:
        date_string = "0860-1-1"
        result = _parse_single_date(date_string, Era.CE)
        expected = SingleDate(era=Era.CE, year=860, month=1, day=1)
        assert result == expected

    def test_fewer_digits_in_year(self) -> None:
        date_string = "860-1-1"
        result = _parse_single_date(date_string, Era.CE)
        expected = SingleDate(era=Era.CE, year=860, month=1, day=1)
        assert result == expected

    def test_do_not_allow_more_digits_in_year(self) -> None:
        date_string = "08600-1-1"
        with pytest.raises(BaseError):
            _parse_single_date(date_string, Era.CE)

    def test_do_not_allow_more_digits_in_month(self) -> None:
        date_string = "0860-100-1"
        with pytest.raises(BaseError):
            _parse_single_date(date_string, Era.CE)

    def test_do_not_allow_more_digits_in_day(self) -> None:
        date_string = "0860-1-100"
        with pytest.raises(BaseError):
            _parse_single_date(date_string, Era.CE)

    def test_do_not_allow_non_date_string(self) -> None:
        date_string = "hello"
        with pytest.raises(BaseError):
            _parse_single_date(date_string, Era.CE)

    def test_do_not_allow_different_date_formats(self) -> None:
        date_string = "01.01.2020"
        with pytest.raises(BaseError):
            _parse_single_date(date_string, Era.CE)


if __name__ == "__main__":
    pytest.main([__file__])



from typing import Any

from dsp_tools.xmllib.models.user_enums import NewlineReplacement


def convert_to_bool_string(value: Any) -> str:
    """
    Turns a value into a bool string, suitable for an XML.

    Args:
        value: value to transform

    Returns:
        'true' or 'false' if it is a known value,
        else it returns the original value as a string.
    """
    str_val = str(value).lower().strip()
    if str_val in ("false", "0", "0.0", "no"):
        return "false"
    elif str_val in ("true", "1", "1.0", "yes"):
        return "true"
    return str(value)


def replace_newlines_with_tags(text: str, converter_option: NewlineReplacement) -> str:
    """
    Converts the newlines in a string to XML tags.
    The type of tags is specified through the converter_option enum.

    Args:
        text: string to convert
        converter_option: tag options

    Returns:
        string with replaced values

    Raises:
        InputError: If an invalid conversion option is given
    """
    match converter_option:
        case NewlineReplacement.NONE:
            return text
        case NewlineReplacement.LINEBREAK:
            return replace_newlines_with_br_tags(text)
        case NewlineReplacement.PARAGRAPH:
            return replace_newlines_with_paragraph_tags(text)


def replace_newlines_with_paragraph_tags(text: str) -> str:
    """
    Replaces '\\n' with `<p>text</p>`

    Args:
        text: string to be formatted

    Returns:
        formatted string with tags
    """
    splt = [x for x in text.split("\n") if x != ""]
    formatted = [f"<p>{x}</p>" for x in splt]
    return "".join(formatted)


def replace_newlines_with_br_tags(text: str) -> str:
    """
    Replaces '\\n' with `<br/>`

    Args:
        text: string to be formatted

    Returns:
        formatted string with tags
    """
    return text.replace("\n", "<br/>")



from pathlib import Path

import pytest
from rdflib import Graph

from dsp_tools.commands.validate_data.models.input_problems import MaxCardinalityViolation
from dsp_tools.commands.validate_data.models.input_problems import MinCardinalityViolation
from dsp_tools.commands.validate_data.models.input_problems import NonExistentCardinalityViolation
from dsp_tools.commands.validate_data.models.validation import ValidationReports
from dsp_tools.commands.validate_data.reformat_validaton_result import reformat_validation_graph
from dsp_tools.commands.validate_data.validate_data import _get_data_info_from_file

LOCAL_API = "http://0.0.0.0:3333"


@pytest.fixture
def data_cardinality_violation() -> Graph:
    data, _ = _get_data_info_from_file(Path("testdata/validate-data/data/cardinality_violation.xml"), LOCAL_API)
    return data.make_graph()


@pytest.fixture
def result_cardinality_violation() -> Graph:
    g = Graph()
    g.parse("testdata/validate-data/validation_results/cardinality_violation_result.ttl")
    return g


def test_reformat_validation_graph(result_cardinality_violation: Graph, data_cardinality_violation: Graph) -> None:
    val_rep = ValidationReports(
        conforms=False,
        content_validation=None,
        cardinality_validation=result_cardinality_violation,
        shacl_graphs=Graph(),
        data_graph=data_cardinality_violation,
    )

    result = reformat_validation_graph(val_rep)
    assert not result.unexpected_results
    assert len(result.problems) == 4
    sorted_problems = sorted(result.problems, key=lambda x: x.sort_value())
    val_one = sorted_problems[0]
    assert isinstance(val_one, MinCardinalityViolation)
    assert val_one.res_id == "id_card_one"
    val_one = sorted_problems[1]
    assert isinstance(val_one, MaxCardinalityViolation)
    assert val_one.res_id == "id_max_card"
    val_one = sorted_problems[2]
    assert isinstance(val_one, MinCardinalityViolation)
    assert val_one.res_id == "id_min_card"
    val_one = sorted_problems[3]
    assert isinstance(val_one, NonExistentCardinalityViolation)
    assert val_one.res_id == "id_closed_constraint"


if __name__ == "__main__":
    pytest.main([__file__])



import pytest
import regex
from lxml import etree

from dsp_tools.models.custom_warnings import DspToolsUserWarning
from dsp_tools.utils.xml_validation import _find_all_text_props_with_multiple_encodings
from dsp_tools.utils.xml_validation import _find_xml_tags_in_simple_text_elements
from dsp_tools.utils.xml_validation import _get_all_ids_and_encodings_from_root
from dsp_tools.utils.xml_validation import _get_encodings_from_one_property
from dsp_tools.utils.xml_validation import _get_encodings_from_one_resource
from dsp_tools.utils.xml_validation import check_if_only_one_encoding_is_used_per_prop_in_root
from dsp_tools.utils.xml_validation_models import TextValueData


class TestFindXMLTagsInUTF8:
    def test_find_xml_tags_in_simple_text_elements_all_good(self, recwarn: pytest.WarningsRecorder) -> None:
        allowed_html_escapes = [
            "(&lt;2cm) (&gt;10cm)",
            "text &lt; text/&gt;",
            "text &lt; text&gt; &amp; text",
            "text &lt;text text &gt; text",
            'text &lt; text text="text"&gt; text',
            'text &lt;text text="text" &gt; text',
        ]
        utf8_texts_with_allowed_html_escapes = [
            f"""
            <knora shortcode="4123" default-ontology="testonto">
                <resource label="label" restype=":restype" id="id">
                    <text-prop name=":name">
                        <text encoding="utf8">{txt}</text>
                    </text-prop>
                </resource>
            </knora>
            """
            for txt in allowed_html_escapes
        ]
        for xml in utf8_texts_with_allowed_html_escapes:
            _find_xml_tags_in_simple_text_elements(etree.fromstring(xml))
        assert len(recwarn) == 0

    def test_find_xml_tags_in_simple_text_elements_forbidden_escapes(self) -> None:
        test_ele = etree.fromstring(
            """
            <knora shortcode="4123" default-ontology="testonto">
                <resource label="label" restype=":restype" id="id">
                    <text-prop name=":name">
                        <text encoding="utf8">&lt;tag s="t"&gt;</text>
                    </text-prop>
                </resource>
            </knora>
            """
        )
        expected_msg = regex.escape(
            "Angular brackets in the format of <text> were found in text properties with encoding=utf8.\n"
            "Please note that these will not be recognised as formatting in the text field, "
            "but will be displayed as-is.\n"
            "The following resources of your XML file contain angular brackets:\n"
            "    - line 5: resource 'id', property ':name'"
        )
        with pytest.warns(DspToolsUserWarning, match=expected_msg):
            _find_xml_tags_in_simple_text_elements(test_ele)

    def test_find_xml_tags_in_simple_text_elements_forbidden_escapes_two(self) -> None:
        test_ele = etree.fromstring(
            """
            <knora shortcode="4123" default-ontology="testonto">
                <resource label="label" restype=":restype" id="id">
                    <text-prop name=":propName">
                        <text encoding="utf8">&lt;em&gt;text&lt;/em&gt;</text>
                    </text-prop>
                </resource>
            </knora>
            """
        )
        expected_msg = regex.escape(
            "Angular brackets in the format of <text> were found in text properties with encoding=utf8.\n"
            "Please note that these will not be recognised as formatting in the text field, "
            "but will be displayed as-is.\n"
            "The following resources of your XML file contain angular brackets:\n"
            "    - line 5: resource 'id', property ':propName'"
        )

        with pytest.warns(DspToolsUserWarning, match=expected_msg):
            _find_xml_tags_in_simple_text_elements(test_ele)


def test_find_all_text_props_with_multiple_encodings_problems() -> None:
    test_props = [TextValueData("problem_id", "problem_prop", {"xml", "utf8"}), TextValueData("", "", {"utf8"})]
    problem = _find_all_text_props_with_multiple_encodings(test_props)[0]
    assert problem.resource_id == "problem_id"
    assert problem.property_name == "problem_prop"
    assert problem.encoding == {"xml", "utf8"}


def test_find_all_text_props_with_multiple_encodings_good() -> None:
    test_props = [TextValueData("", "", {"xml"}), TextValueData("", "", {"utf8"})]
    problem = _find_all_text_props_with_multiple_encodings(test_props)
    assert not problem


def test_get_all_ids_prop_encoding_from_root_no_text() -> None:
    test_ele = etree.fromstring(
        """<knora>
                <resource label="resA" restype=":TestThing1" id="resA" permissions="res-default">
                    <resptr-prop name=":hasResource1">
                        <resptr permissions="prop-default">resB</resptr>
                    </resptr-prop>
                </resource>
                <resource label="resB" restype=":TestThing2" id="resB" permissions="res-default">
                </resource>
                <resource label="resC" restype=":TestThing2" id="resC" permissions="res-default">
                    <resptr-prop name=":hasResource2">
                        <resptr permissions="prop-default">resB</resptr>
                    </resptr-prop>
                    <resptr-prop name=":hasResource3">
                        <resptr permissions="prop-default">resA</resptr>
                    </resptr-prop>
                </resource>
        </knora>"""
    )
    res = _get_all_ids_and_encodings_from_root(test_ele)
    assert not res


def test_get_all_ids_prop_encoding_from_root_with_text() -> None:
    test_ele = etree.fromstring(
        """<knora>
                <resource label="First Testthing"
                      restype=":TestThing"
                      id="test_thing_1"
                      permissions="res-default">
                    <uri-prop name=":hasUri">
                        <uri permissions="prop-default">https://dasch.swiss</uri>
                    </uri-prop>
                    <text-prop name=":hasRichtext">
                        <text encoding="xml">Text</text>
                    </text-prop>
                </resource>
                <resource label="resB" restype=":TestThing2" id="resB" permissions="res-default">
                    <text-prop name=":hasSimpleText">
                        <text encoding="utf8">Text</text>
                    </text-prop>
                </resource>
                <resource label="resC" restype=":TestThing2" id="resC" permissions="res-default">
                    <resptr-prop name=":hasResource2">
                        <resptr permissions="prop-default">resB</resptr>
                    </resptr-prop>
                    <text-prop name=":hasSimpleText">
                        <text encoding="utf8">Text</text>
                    </text-prop>
                </resource>
                <resource label="resD" restype=":TestThing2" id="resD" permissions="res-default">
                    <resptr-prop name=":hasResource2">
                        <resptr permissions="prop-default">resB</resptr>
                    </resptr-prop>
                </resource>
        </knora>"""
    )
    res = _get_all_ids_and_encodings_from_root(test_ele)
    assert res[0].resource_id == "test_thing_1"
    assert res[0].property_name == ":hasRichtext"
    assert res[0].encoding == {"xml"}
    assert res[1].resource_id == "resB"
    assert res[1].property_name == ":hasSimpleText"
    assert res[1].encoding == {"utf8"}
    assert res[2].resource_id == "resC"
    assert res[2].property_name == ":hasSimpleText"
    assert res[2].encoding == {"utf8"}


class TestGetEncodingsOneResource:
    def test_no_text(self) -> None:
        test_props = etree.fromstring(
            """
            <resource label="First Testthing"
                  restype=":TestThing"
                  id="test_thing_1"
                  permissions="res-default">
                <uri-prop name=":hasUri">
                    <uri permissions="prop-default">https://dasch.swiss</uri>
                </uri-prop>
                <boolean-prop name=":hasBoolean">
                    <boolean permissions="prop-default">true</boolean>
                </boolean-prop>
            </resource>
            """
        )
        res = _get_encodings_from_one_resource(test_props)
        assert not res

    def test_one_text_prop(self) -> None:
        test_props = etree.fromstring(
            """
            <resource label="First Testthing"
                  restype=":TestThing"
                  id="test_thing_1"
                  permissions="res-default">
                <uri-prop name=":hasUri">
                    <uri permissions="prop-default">https://dasch.swiss</uri>
                </uri-prop>
                <boolean-prop name=":hasBoolean">
                    <boolean permissions="prop-default">true</boolean>
                </boolean-prop>
                <text-prop name=":hasRichtext">
                    <text encoding="utf8">Text</text>
                </text-prop>
            </resource>
            """
        )
        res = _get_encodings_from_one_resource(test_props)[0]
        assert res.resource_id == "test_thing_1"
        assert res.property_name == ":hasRichtext"
        assert res.encoding == {"utf8"}

    def test_two_text_prop(self) -> None:
        test_props = etree.fromstring(
            """
            <resource label="First Testthing"
                  restype=":TestThing"
                  id="test_thing_1"
                  permissions="res-default">
                <uri-prop name=":hasUri">
                    <uri permissions="prop-default">https://dasch.swiss</uri>
                </uri-prop>
                <boolean-prop name=":hasBoolean">
                    <boolean permissions="prop-default">true</boolean>
                </boolean-prop>
                <text-prop name=":hasRichtext">
                    <text encoding="xml">Text</text>
                </text-prop>
                <text-prop name=":hasSimpleText">
                    <text encoding="utf8">Text</text>
                    <text encoding="utf8">Text</text>
                </text-prop>
            </resource>
            """
        )
        res = _get_encodings_from_one_resource(test_props)
        assert res[0].resource_id == "test_thing_1"
        assert res[0].property_name == ":hasRichtext"
        assert res[0].encoding == {"xml"}
        assert res[1].resource_id == "test_thing_1"
        assert res[1].property_name == ":hasSimpleText"
        assert res[1].encoding == {"utf8"}


class TestGetEncodingOneProperty:
    def test_richtext_several_text_ele(self) -> None:
        test_prop = etree.fromstring(
            """
            <text-prop name=":hasRichtext">
                <text encoding="xml">&lt;</text>
                <text encoding="xml" permissions="prop-default">
                    This text contains links to all resources:
                    <a class="salsah-link" href="IRI:test_thing_0:IRI">test_thing_0</a>
                </text>
                <text encoding="xml">Text with an external link: <a href="https://www.google.com/">Google</a></text>
            </text-prop>
            """
        )
        res_info = _get_encodings_from_one_property("id", test_prop)
        assert res_info.resource_id == "id"
        assert res_info.property_name == ":hasRichtext"
        assert res_info.encoding == {"xml"}

    def test_simple_several_text_ele(self) -> None:
        test_prop = etree.fromstring(
            """
            <text-prop name=":hasRichtext">
                <text encoding="utf8">Text</text>
                <text encoding="utf8">Text</text>
            </text-prop>
            """
        )
        res_info = _get_encodings_from_one_property("id", test_prop)
        assert res_info.resource_id == "id"
        assert res_info.property_name == ":hasRichtext"
        assert res_info.encoding == {"utf8"}

    def test_simple_one_text_ele(self) -> None:
        test_prop = etree.fromstring(
            """
            <text-prop name=":hasRichtext">
                <text encoding="utf8">Text</text>
            </text-prop>
            """
        )
        res_info = _get_encodings_from_one_property("id", test_prop)
        assert res_info.resource_id == "id"
        assert res_info.property_name == ":hasRichtext"
        assert res_info.encoding == {"utf8"}


def test_check_if_only_one_encoding_is_used_in_xml_good() -> None:
    test_ele = etree.fromstring(
        """<knora>
                <resource label="First Testthing"
                      restype=":TestThing"
                      id="test_thing_1"
                      permissions="res-default">
                    <uri-prop name=":hasUri">
                        <uri permissions="prop-default">https://dasch.swiss</uri>
                    </uri-prop>
                    <text-prop name=":hasRichtext">
                        <text encoding="xml">Text</text>
                    </text-prop>
                </resource>
                <resource label="resB" restype=":TestThing2" id="resB" permissions="res-default">
                    <text-prop name=":hasSimpleText">
                        <text encoding="utf8">Text</text>
                    </text-prop>
                </resource>
                <resource label="resC" restype=":TestThing2" id="resC" permissions="res-default">
                    <resptr-prop name=":hasResource2">
                        <resptr permissions="prop-default">resB</resptr>
                    </resptr-prop>
                    <text-prop name=":hasSimpleText">
                        <text encoding="utf8">Text 1</text>
                        <text encoding="utf8">Text 2</text>
                    </text-prop>
                </resource>
                <resource label="resD" restype=":TestThing2" id="resD" permissions="res-default">
                    <resptr-prop name=":hasResource2">
                        <resptr permissions="prop-default">resB</resptr>
                    </resptr-prop>
                </resource>
        </knora>"""
    )
    res = check_if_only_one_encoding_is_used_per_prop_in_root(test_ele)
    assert not res


def test_check_if_only_one_encoding_is_used_in_xml_problem() -> None:
    test_ele = etree.fromstring(
        """<knora>
                <resource label="First Testthing"
                      restype=":TestThing"
                      id="test_thing_1"
                      permissions="res-default">
                    <uri-prop name=":hasUri">
                        <uri permissions="prop-default">https://dasch.swiss</uri>
                    </uri-prop>
                    <text-prop name=":hasRichtext">
                        <text encoding="xml">Text</text>
                    </text-prop>
                </resource>
                <resource label="resB" restype=":TestThing2" id="resB" permissions="res-default">
                    <text-prop name=":hasSimpleText">
                        <text encoding="utf8">Text</text>
                    </text-prop>
                </resource>
                <resource label="resC" restype=":TestThing2" id="resC" permissions="res-default">
                    <resptr-prop name=":hasResource2">
                        <resptr permissions="prop-default">resB</resptr>
                    </resptr-prop>
                    <text-prop name=":hasSimpleText">
                        <text encoding="utf8">Text 1</text>
                        <text encoding="xml">Text 2</text>
                    </text-prop>
                </resource>
                <resource label="resD" restype=":TestThing2" id="resD" permissions="res-default">
                    <resptr-prop name=":hasResource2">
                        <resptr permissions="prop-default">resB</resptr>
                    </resptr-prop>
                </resource>
        </knora>"""
    )
    res = check_if_only_one_encoding_is_used_per_prop_in_root(test_ele)[0]
    assert res.resource_id == "resC"
    assert res.property_name == ":hasSimpleText"
    assert res.encoding == {"xml", "utf8"}


if __name__ == "__main__":
    pytest.main([__file__])



import json
from typing import Any

import pandas as pd
import regex


def is_bool_like(value: Any) -> bool:
    """
    Checks if a value is a bool or can be converted into a bool.

    Args:
        value: value to check

    Returns:
        True if it conforms
    """
    value = str(value).lower().strip()
    if value in ("false", "0", "0.0", "no"):
        return True
    elif value in ("true", "1", "1.0", "yes"):
        return True
    return False


def is_color(value: Any) -> bool:
    """
    Checks if a value is a color value.

    Args:
        value: value to check

    Returns:
        True if it conforms
    """

    return bool(regex.search(r"^#[0-9a-f]{6}$", str(value).strip(), flags=regex.IGNORECASE))


def is_date(value: Any) -> bool:
    """
    Checks if a value is a date value.

    Args:
        value: value to check

    Returns:
        True if it conforms
    """

    calendar_optional = r"((GREGORIAN|JULIAN|ISLAMIC):)?"
    first_era_optional = r"((CE|BCE|BC|AD):)?"
    second_area_optional = r"(:(CE|BCE|BC|AD))?"
    date = r"\d{1,4}(?:-\d{1,2}){0,2}"
    date_mandatory = rf"({date})"
    date_optional = rf"(:{date})?"
    full_date_pattern = (
        rf"^{calendar_optional}{first_era_optional}{date_mandatory}{second_area_optional}{date_optional}$"
    )
    return bool(regex.search(full_date_pattern, str(value)))


def is_geoname(value: Any) -> bool:
    """
    Checks if a value is a geoname value.

    Args:
        value: value to check

    Returns:
        True if it conforms
    """
    return is_integer(value)


def is_decimal(value: Any) -> bool:
    """
    Checks if a value is a float, an integer, or a string which can be converted into a float.

    Args:
        value: value to check

    Returns:
        True if conforms to the above-mentioned criteria.
    """
    if pd.isna(value):
        return False

    match value:
        case bool():
            return False
        case int() | float():
            return True
    try:
        float(value)
        return True
    except ValueError:
        return False


def is_integer(value: Any) -> bool:
    """
    Checks if a value is an integer or a string which can be converted into an integer.

    Args:
        value: value to check

    Returns:
        True if conforms to the above-mentioned criteria.
    """
    match value:
        case bool():
            return False
        case int():
            return True
        case str():
            return bool(regex.search(r"^\d+$", value))
        case _:
            return False


def is_string_like(value: Any) -> bool:
    """
    Checks if a value is a string.

    Args:
        value: value to check

    Returns:
        True if it is a string
    """
    if pd.isna(value):
        return False
    value = str(value).strip()
    if len(value) == 0:
        return False
    return bool(regex.search(r"\S", value, flags=regex.UNICODE))


def is_timestamp(value: Any) -> bool:
    """
    Checks if a value is a valid timestamp.

    Args:
        value: value to check

    Returns:
        True if it conforms
    """
    validation_regex = r"^\d{4}-[0-1]\d-[0-3]\dT[0-2]\d:[0-5]\d:[0-5]\d(.\d{1,12})?(Z|[+-][0-1]\d:[0-5]\d)$"
    return bool(regex.search(validation_regex, str(value)))


def find_geometry_problem(value: Any) -> str:
    """
    Validates if a value is a valid geometry object.

    Args:
        value: geometry object

    Returns:
        string with the validation message if it fails
    """
    msg = ""
    try:
        value_as_dict = json.loads(str(value))
        if value_as_dict["type"] not in ["rectangle", "circle", "polygon"]:
            msg += "\nThe 'type' of the JSON geometry object must be 'rectangle', 'circle', or 'polygon'."

        if not isinstance(value_as_dict["points"], list):
            msg += "\nThe 'points' of the JSON geometry object must be a list of points."
    except (json.JSONDecodeError, TypeError, IndexError, KeyError, AssertionError):
        msg += f"\n'{value}' is not a valid JSON geometry object."
    return msg


def is_dsp_iri(value: Any) -> bool:
    """Checks if a value is a valid internal dsp IRI"""
    return bool(regex.search(r"^http://rdfh\.ch/\d{4}/", str(value)))


def is_dsp_ark(value: Any) -> bool:
    """Checks if a value is a valid ARK"""
    return bool(regex.search(r"^ark:/", str(value)))



from __future__ import annotations

from dataclasses import dataclass

from rdflib import Graph
from rdflib.term import Node


@dataclass
class RDFGraphs:
    data: Graph
    ontos: Graph
    cardinality_shapes: Graph
    content_shapes: Graph

    def get_cardinality_data_str(self) -> str:
        return self.data.serialize(format="ttl")

    def get_cardinality_shacl_str(self) -> str:
        return self.cardinality_shapes.serialize(format="ttl")

    def get_content_data_str(self) -> str:
        g = self.data + self.ontos
        return g.serialize(format="ttl")

    def get_content_shacl_str(self) -> str:
        return self.content_shapes.serialize(format="ttl")


@dataclass
class ValidationReports:
    conforms: bool
    content_validation: Graph | None
    cardinality_validation: Graph | None
    shacl_graphs: Graph
    data_graph: Graph


@dataclass
class UnexpectedComponent:
    component_type: str


@dataclass
class CardinalityValidationResult:
    source_constraint_component: Node
    res_iri: Node
    res_class: Node
    property: Node
    results_message: str


@dataclass
class ContentValidationResult:
    source_constraint_component: Node
    res_iri: Node
    res_class: Node
    property: Node
    results_message: str
    detail_bn_component: Node
    value: str | None
    value_type: Node | None = None
